/*!
 ***********************************************************************
 *  @file	     adc.c
 *  @par Project   : c't Mikrocontroller-im-LAN --> Wetterstation
 *  @par Module    : ADC-Wandler
 *
 *  @brief	Funktionen zum einlesen analoger Werte
 *
 *  @author Benjamin Benz (bbe@heise.de)
 *  @author Markus Thiele (Chewy.F5@gmx.de)
 *  @version 2.0
 *	@date 08.01.2005
 ***********************************************************************
 *
 *  @par History   :
 *  @verbatim
 *  Version | Date       | Author    	      | Change Description
 *  1.0     | 24.06.2004 | Benz               | First Version, Demo Source Code
 *  2.0     | 08.01.2005 | Thiele             | First Documentated Version of Wetterstation
 *  @endverbatim
 *
 ************************************************************************/

/*! \def T1m
 *  \brief Temperatur Sensor 1 / PT100 Faktor
 *
 *  Definiert den 10 fachen Faktor von dTemperatur / dMesswert
 ****************************************************************************************/
/*! \def T1b
 *  \brief Temperatur Sensor 1 / PT100 Offset
 *
 *  Definiert den 10 fachen Offset der Temperatur
 ****************************************************************************************/
/*! \def T2m
 *  \brief Temperatur Sensor 2 / PT100 Faktor
 *
 *  Definiert den 10 fachen Faktor von dTemperatur / dMesswert
 ****************************************************************************************/
/*! \def T2b
 *  \brief Temperatur Sensor 2 / PT100 Offset
 *
 *  Definiert den 10 fachen Offset der Temperatur
 ****************************************************************************************/
/*! \def D1m
 *  \brief Drucksensor / MPX 4115A Faktor
 *
 *  Definiert den 10 fachen Faktor von dDruck / dMesswert
 ****************************************************************************************/
/*! \def D1b
 *  \brief Drucksensor / MPX 4115A Faktor
 *
 *  Definiert den 10 fachen Offset des Drucks
 ****************************************************************************************/

#include <avr/io.h>
#include <avr/delay.h>
#include <avr/sleep.h>
#include <stdlib.h>
#include "adc.h"
#include "config.h"

// Globale ADC Variabeln
//define T1L 	48.7122				// PT 1 Wert fuer 0 Grad
//define T1M 	1015.888			// Wert fuer 100 Grad
//define T2L 	114.1357			// PT 2 Wert ...
//define T2M 	1019.244
//define DRUCK_SPANNUNG 4.93 		// Versorgungspannung am Drucksensor
//define DRUCK_ERROR 	0.053244	// Offsettfehler
//define SPANNUNG_A		2.56		// Spannung am AVCC 4,92 Referenz = 2,56Vdes AVRs
/* Temp.Sensoren vom TYP PT100 , Drucksensor Motorola MPX4115A */
#define T1m		1.033938f                                   // Temp1 Faktor 100 / ( T1M - T1L ) *10
#define T1b		100.3654f                                   // Temp1 Konst. T1L * 100 / ( T1M - T1L ) + 5  *10 Da der Tempfhler immer 5 Grad zuviel anzeigt
#define T2m		1.10484f                                    // Temp2 Faktor 100 / ( T2M - T2L ) *10
#define T2b		126.1f                                      // Temp2 Konst. T2L * 100 / ( T2M - T2L ) *10
#define D1m		2.09602f                                    // Druck Faktor *10
#define D1b		8818.81f                                    // Druck Konstante *10


int read_adc(unsigned char channel){                        //Liest einen analogen kanal aus
                                                            //Der Parameter ist die Nummer des Pins (0x00 fr PA0; 0x01 fr PA1, ..)
	unsigned int result = 0x00;
	unsigned char old_DDRA= DDRA;                           // Alte Einstellungen sichern
	unsigned char old_PORTA= PORTA;
	unsigned char old_admux  = ADMUX;

	// Sync Zusatz um zu verhindern das durch mehrmaligen aufruf ADC-Werte vertaucht werden
	unsigned char old_adcsra = ADCSRA;                      // Merke Dir die alten Einstellungen
	unsigned char merker;                                   // genutzt wenn der ADC schon luft
	if ((ADCSRA & (1<<ADSC)) != 0)                          // Luft der ADC schon?
	  {
		//old_adcsra = (ADCSRA);
		while ( (ADCSRA & (1<<ADSC)) != 0){}                // Warte bis die Alte Wandlung zuende ist
		merker=1;                                           // Merker an
	  }
	else merker=0;                                          // Nein, dann merke Dir nichts
	// Sync Zusatz Ende

	DDRA &= ~ (1<<channel);                                 // Pin als input
	PORTA &= ~ (1<<channel);                                // Alle Pullups aus.

	ADMUX = 0xC0;                                           // interne Refernzspannung, rechts Ausrichtung (0x80+0x40)
	ADMUX |= channel;                                       // Und jetzt Kanal whlen
	asm("nop");                                             // Erstmal eine kurze Pause machen bevor man was tut

	ADCSRA=
		(1<<ADPS2) + (1<<ADPS1)+ (1<<ADPS0)+                // prescale faktor = 128 ADC luft mit 14,7456MHz / 128 = 115,2 kHz
		(1 << ADEN);                                        // ADC an

	//set_sleep_mode (SLEEP_MODE_ADC) ;                     // Problem: siehe unten

	asm("nop");                                             // Eile mit Weile, lass den ADC erstmal anlaufen
	//sleep_mode();                                         // \todo sleep_mode Strt beim Senden der Daten!
	ADCSRA |= (1 << ADSC);                                  // Beginne mit der Konvertierung


	while ( (ADCSRA & (1<<ADSC)) != 0){}                    // Maskiere ADSC und Warten bis konvertierung beendet
	asm("nop");
	result= ADCL + (ADCH <<8);                              // Ergebnis zusammenbauen

	DDRA= old_DDRA;                                         // wiederherstellen
	PORTA= old_PORTA;
	ADMUX = old_admux;                                      // Lade alten Mux

	// Sync Zusatz
	if (merker != 0)                                        // Wurde oben festgestellt das der ADC beim start schon lief?
	  {
	   ADCSRA = old_adcsra & (1 << ADSC);                   // Lade und starte mit alten einstellungen
	  }
	// Sync Ende
	return result;
}

short read_zahl_adc(unsigned char channel) {
	switch (channel) {
		default: // PT1
				return (( ((float)read_adc(channel)) * ( T1m )) - ( T1b )) ;
				break;
		case 1: // PT2
				return (( ((float)read_adc(channel)) * ( T2m )) - ( T2b )) ;
				break;
		case 2: // LDR
				return ( (float)read_adc(channel)*10.0f );
				break;
		case 3: // Druck
				return (( ((float)read_adc(5)) * ( D1m )) + ( D1b )) ;
				break;
		#if ( MAX_CHANNEL > 4 )
		case 4: // Feuchte, noch ohne Wert da es noch keinen gibt.
				return ((float)read_adc(6) *10.0f);
				break;
		#endif
	}
	return (0);
}

void read_string_adc(unsigned char channel,unsigned char *data){
	switch(channel) {
		default: // PT
				dtostrf((float)read_zahl_adc(channel)/10.0f,5,1,data);
				data[5]=0xDF;
				data[6]='C';
				data[7]=0x00;
				break;
		case 2: // LDR
				itoa((int)read_zahl_adc(channel)/10,data,10);
				break;
		case 3: // Druck
				dtostrf((float)read_zahl_adc(channel)/10.0f,6,1,data);
				data[6]='m';
				data[7]='B';
				data[8]='a';
				data[9]='r';
				data[10]=0x00;
				break;
		#if ( MAX_CHANNEL > 4 )
		case 4: // Feuchte
				dtostrf((float)read_zahl_adc(channel)/10.0f,5,1,data);
				data[5]='%';
				data[6]='r';
				data[7]='.';
				data[8]='F';
				data[9]='.';
				data[10]=0x00;
				break;
		#endif
	}
}